from passthesalt.exceptions import ConflictingTimestamps, PassTheSaltError, UnauthorizedAccess


def test_errors():
    e = PassTheSaltError('msg')
    assert e.message == 'msg'
    assert str(e) == 'msg'
    assert repr(e) == "passthesalt.exceptions.PassTheSaltError('msg')"

    e = UnauthorizedAccess('msg')
    assert e.message == 'msg'
    assert str(e) == 'msg'
    assert repr(e) == "passthesalt.exceptions.UnauthorizedAccess('msg', code=401)"
    assert e.code == 401

    e = ConflictingTimestamps('msg')
    assert e.message == 'msg'
    assert str(e) == 'msg'
    assert repr(e) == "passthesalt.exceptions.ConflictingTimestamps('msg', code=409)"
    assert e.code == 409



"""
A command line interface for PassTheSalt.
"""

import datetime
import json
import os
import subprocess
import sys
from functools import wraps

import click
import pyperclip
import serde
import validators
from click import confirm, echo
from tabulate import tabulate

from passthesalt import Algorithm, Encrypted, Generatable, Login, Master, PassTheSalt, Secret
from passthesalt.exceptions import LabelError, PassTheSaltError
from passthesalt.remote import Stow


class DomainParamType(click.ParamType):
    """
    A click Domain parameter type.
    """

    name = 'domain'

    def convert(self, value, param, ctx):
        """
        Validate the entered value is a domain.
        """
        if not validators.domain(value):
            self.fail(f'{value} is not a valid domain', param, ctx)

        return value


class UrlParamType(click.ParamType):
    """
    A click URL parameter type.
    """

    name = 'url'

    def convert(self, value, param, ctx):
        """
        Validate the entered value is a URL.
        """
        if not validators.url(value):
            self.fail(f'{value} is not a valid URL', param, ctx)

        return value


DOMAIN = DomainParamType()
URL = UrlParamType()
DEFAULT_PATH = os.path.expanduser('~/.passthesalt')
DEFAULT_REMOTE_PATH = os.path.expanduser('~/.passthesalt.remote')


def bail(message):
    """
    Abort the CLI with a message.
    """
    raise click.ClickException(message)


def prompt(*args, none_if_default=False, default=None, **kwargs):
    """
    Prompts a user for input.

    This is simply a wrapper around click.prompt() to provide the option to
    return None in the case that the prompt returns the default.

    Args:
        *args: positional arguments to pass directly to `click.prompt()`.
        none_if_default (bool): if True then None will be returned in the case
            that the returned value is the same as the default.
        default: the input to default to if the user inputs nothing. This is
            passed directly to `click.prompt()`.
        **kwargs: keyword arguments to pass directly to `click.prompt()`.

    Returns:
        the user input.
    """
    result = click.prompt(*args, default=default, **kwargs)

    if not none_if_default or result != default:
        return result


def handle_passthesalt_errors(f):
    """
    Translate PassTheSaltErrorErrors to ClickExceptions.

    Args:
        f (function): the function to decorate.

    Raises:
        click.ClickException: when the function raises a PassTheSaltError.

    Returns:
        function: the decorated function.
    """
    @wraps(f)
    def decorated_function(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except PassTheSaltError as e:
            bail(e.message)

    return decorated_function


def clear_clipboard(timeout):
    """
    Clear the clipboard after a timeout.

    Args:
        timeout (int): the timeout.
    """
    code = f"import pyperclip, time; time.sleep({timeout}); pyperclip.copy('');"
    command = f'{sys.executable} -c "{code}"'
    subprocess.Popen(
        command,
        stdin=subprocess.DEVNULL,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
        shell=True
    )


def copy_to_clipboard(text, timeout=None):
    """
    Copy the given text to clipboard.

    Args:
        text (str): the text to copy to the clipboard.
        timeout (int): clear the clipboard after this amount of seconds.
    """
    pyperclip.copy(text)

    if timeout:
        clear_clipboard(timeout)


def ask_user_for_master(pts):
    """
    Prompt the user for the master password. Bail after 3 attempts.

    Args:
        pts (PassTheSalt): the PassTheSalt instance calling this function.

    Returns:
        str: the master password.
    """
    for i in range(3):
        master = prompt('Enter master password', hide_input=True)

        if pts.config.master.is_valid(master):
            return master

        echo('Error: invalid master password', err=True)

    raise bail('three incorrect attempts')


def ask_user_for_auth(remote):
    """
    Prompt the user for the username and password for the remote store.

    Args:
        remote (Remote): the remote instance.

    Returns:
        (str, str): the username and password.
    """
    echo('Authentication is required')
    name = prompt('Enter username')
    password = prompt('Enter password', hide_input=True)

    return name, password


def read_or_init_remote(path):
    """
    Read the Remote from the given path or initialize a new Remote.

    Args:
        path (str): the path to the Remote.

    Returns:
        Remote: the initialized or read Remote.
    """
    try:
        remote = Stow.from_path(path).with_auth(ask_user_for_auth)
    except OSError:
        echo('Initializing PassTheSalt remote configuration ...')
        remote = Stow(
            location=prompt('Enter location URL', type=URL),
            token_location=prompt('Enter token renew URL', type=URL)
        ).with_auth(ask_user_for_auth)

    return remote


def pts_ls_(pts, label=None, kind=None, header=True, verbose=1):
    """
    List the secrets in a PassTheSalt store.

    Args:
        pts (PassTheSalt): the PassTheSalt store.
        label (str): a label to filter on.
        kind (str): the kind to filter on (generatable or encrypted)
        header (bool): whether to print the table header.
        verbose (int): the verbosity count that determines how much information
            for each secret is displayed.
    """
    labels = pts.labels(pattern=label)

    if not labels:
        echo('No stored secrets', err=True)
    else:
        secrets = []

        for label in sorted(labels):
            secret = pts.get(label)

            if not kind or secret.kind == kind:
                secrets.append(secret.display()[:verbose + 1])

        if header:
            kwargs = {
                'headers': ('Label', 'Kind', 'Modified', 'Salt')[:verbose + 1]
            }
        else:
            kwargs = {
                'tablefmt': 'plain',
            }

        echo(tabulate(secrets, **kwargs))


@click.group(
    context_settings={'help_option_names': ['-h', '--help']},
    invoke_without_command=True
)
@click.version_option(
    None,
    '-v', '--version',
    prog_name='passthesalt',
    message='%(prog)s %(version)s'
)
@click.option(
    '--path', '-p',
    type=click.Path(),
    default=DEFAULT_PATH,
    help='The path to the PassTheSalt store.'
)
@click.pass_context
@handle_passthesalt_errors
def cli(ctx, path):
    """
    \b
        ____                 ________        _____       ____
       / __ \____  _________/_  __/ /_  ___ / ___/____ _/ / /_
      / /_/ / __ `/ ___/ ___// / / __ \/ _ \\\\__ \/ __ `/ / __/
     / ____/ /_/ (__  |__  )/ / / / / /  __/__/ / /_/ / / /_
    /_/    \__,_/____/____//_/ /_/ /_/\___/____/\__,_/_/\__/

    A deterministic password generation and password storage system.
    """
    pts = None

    try:
        pts = PassTheSalt.from_path(path).with_master(ask_user_for_master).with_path(path)
    except OSError:
        pass

    # The only time pts is allowed to be None is if we are pulling a store.
    if ctx.invoked_subcommand != 'pull':
        if pts is None:
            echo('Initializing PassTheSalt ...')
            pts = PassTheSalt().with_master(ask_user_for_master).with_path(path)
            pts.config.owner = prompt('Please enter your name')
            pts.config.master = Master(prompt(
                'Please enter the master password',
                confirmation_prompt=True,
                hide_input=True
            ))
            pts.save()
            echo('Successfully initialized PassTheSalt!')
        elif ctx.invoked_subcommand is None:
            echo(ctx.get_help())

    ctx.obj = pts or path


@cli.command('add')
@click.argument('label', required=False)
@click.option(
    '--type', '-t',
    type=click.Choice(['raw', 'login']),
    default='login',
    help='The type of generated secret.'
)
@click.option(
    '--length', '-l',
    type=int,
    help='The length of the generated secret.'
)
@click.option(
    '--version', '-v',
    type=click.Choice(['0', '1']),
    default='1',
    show_default=True,
    help='The algorithm version to use.'
)
@click.option(
    '--clipboard/--no-clipboard',
    default=True,
    show_default=True,
    help='Whether to copy the secret to the clipboard or print it out.'
)
@click.pass_context
@handle_passthesalt_errors
def pts_add(ctx, label, type, length, version, clipboard):
    """
    Add a secret.

    Add a generatable secret with label LABEL to the PassTheSalt store.
    """
    pts = ctx.obj

    if not label:
        label = prompt('Enter label')

    if pts.contains(label):
        bail(f'{label!r} already exists')

    algorithm = Algorithm(version=int(version), length=length)

    if type == 'login':
        secret = Login(
            domain=prompt('Enter domain name', type=DOMAIN),
            username=prompt('Enter username'),
            iteration=prompt('Enter iteration', default=0),
            algorithm=algorithm
        )
    else:
        secret = Generatable(
            salt=prompt('Enter salt'),
            algorithm=algorithm
        )

    confirm(f'Store {secret.salt!r} as {label!r}?', abort=True)
    pts.add(label, secret)
    pts.save()
    echo(f'Stored {label!r}!')

    if confirm('\nRetrieve secret?'):
        ctx.invoke(pts_get, label=label, clipboard=clipboard)


@cli.command('encrypt')
@click.argument('label', required=False)
@click.option('--secret', '-s', help='The secret to store.')
@click.pass_obj
@handle_passthesalt_errors
def pts_encrypt(pts, label, secret):
    """
    Encrypt a secret.

    Add an encrypted secret with label LABEL to the PassTheSalt store.
    """
    if not label:
        label = prompt('Enter label')

    if pts.contains(label):
        bail(f'{label!r} already exists')

    if not secret:
        secret = prompt('Enter secret', confirmation_prompt=True, hide_input=True)

    pts.add(label, Encrypted(secret))
    pts.save()
    echo(f'Stored {label!r}!')


@cli.command('edit')
@click.argument('label', required=False)
@click.option(
    '--clipboard/--no-clipboard',
    default=True,
    show_default=True,
    help='Whether to copy the secret to the clipboard or print it out.'
)
@click.pass_context
@handle_passthesalt_errors
def pts_edit(ctx, label, clipboard):
    """
    Edit a secret.

    Edit a secret matching the label LABEL. This will open the secret using the
    default EDITOR.
    """
    pts = ctx.obj

    if not label:
        label = prompt('Enter label')

    label = pts.resolve(pattern=label)
    secret = pts.get(label)

    updated = False

    if isinstance(secret, Encrypted):
        raw = secret.get()
        new_raw = click.edit(raw)

        if new_raw:
            if new_raw != raw:
                new_secret = Encrypted(new_raw)
                pts.update(label, new_secret)
                updated = True
            else:
                echo(f'{label!r} was not changed')
        else:
            raise click.Abort()
    else:
        new_toml = click.edit(secret.to_toml())

        if new_toml:
            new_secret = Secret.from_toml(new_toml)

            if new_secret != secret:
                new_secret.touch()  # because Secret.from_toml() doesn't update the modified time.
                pts.update(label, new_secret)
                updated = True
            else:
                echo(f'{label!r} was not changed')
        else:
            raise click.Abort()

    if updated:
        pts.save()
        echo(f'Updated {label!r}!')

        if confirm('\nRetrieve secret?'):
            ctx.invoke(pts_get, label=label, clipboard=clipboard)


@cli.command('get')
@click.argument('label', required=False)
@click.option(
    '--clipboard/--no-clipboard',
    default=True,
    show_default=True,
    help='Whether to copy the secret to the clipboard or print it out.'
)
@click.pass_obj
@handle_passthesalt_errors
def pts_get(pts, label, clipboard):
    """
    Retrieve a secret.

    Get the secret matching the label LABEL.
    """
    if not label:
        label = prompt('Enter label')

    label = pts.resolve(pattern=label)
    secret = pts.get(label).get()

    if clipboard:
        copy_to_clipboard(secret, timeout=20)
        echo('Secret copied to clipboard.')
    else:
        echo(secret)


@cli.command('ls')
@click.argument('label', required=False)
@click.option(
    '--kind', '-k',
    type=click.Choice(['encrypted', 'generatable']),
    help='Filter by type of secret.'
)
@click.option(
    '--header/--no-header',
    default=True,
    help='Whether to display the table header.'
)
@click.option(
    '--verbose', '-v',
    count=True,
    help='Increase amount information displayed.'
)
@click.pass_obj
@handle_passthesalt_errors
def pts_ls(pts, label, kind, header, verbose):
    """
    List the secrets.
    """
    pts_ls_(pts, label=label, kind=kind, header=header, verbose=verbose)


@cli.command('rm')
@click.argument('label', required=False)
@click.option(
    '--regex', '-r',
    is_flag=True,
    help='Match labels using LABEL as a regex pattern.'
)
@click.option(
    '--force', '-f',
    is_flag=True,
    help='Do not ask for confirmation before removing.'
)
@click.pass_obj
@handle_passthesalt_errors
def pts_rm(pts, label, regex, force):
    """
    Remove a secret.

    Completely remove the secret(s) matching the label LABEL from the
    PassTheSalt store. Removing encrypted secrets will require the master
    password.
    """
    if not label:
        label = prompt('Enter label')

    try:
        if regex:
            label = pts.resolve(pattern=label)
    except LabelError:
        labels = pts.labels(pattern=label)

        if not labels:
            raise

        labels_display = ', '.join(repr(label) for label in labels)

        if not force:
            confirm(f'Remove {labels_display}?', abort=True)

        for label in labels:
            pts.remove(label)

        pts.save()
        echo(f'Removed {labels_display}!')
    else:
        if not pts.contains(label):
            bail(f'{label!r} does not exist')

        if not force:
            confirm(f'Remove {label!r}?', abort=True)

        pts.remove(label)
        pts.save()
        echo(f'Removed {label!r}!')


@cli.command('mv')
@click.argument('label')
@click.argument('new-label', metavar='NEW-LABEL')
@click.pass_obj
@handle_passthesalt_errors
def pts_mv(pts, label, new_label):
    """
    Relabel a secret.

    Give the secret matching the label LABEL a new label NEW-LABEL.
    """
    pts.move(label, new_label)
    pts.save()
    echo(f'Renamed {label!r} as {new_label!r}!')


@cli.command('push')
@click.option(
    '--path', '-p',
    type=click.Path(),
    default=DEFAULT_REMOTE_PATH,
    help='The path to the PassTheSalt remote configuration.'
)
@click.option(
    '--force', '-f',
    is_flag=True,
    help='Ignore any conflicts.'
)
@click.pass_obj
@handle_passthesalt_errors
def pts_push(pts, path, force):
    """
    Update the remote store.
    """
    remote = read_or_init_remote(path)
    remote.put(pts, force=force)
    remote.to_path(path)
    echo('Successfully pushed to remote store')


@cli.command('pull')
@click.option(
    '--path', '-p',
    type=click.Path(),
    default=DEFAULT_REMOTE_PATH,
    help='The path to the PassTheSalt remote configuration.'
)
@click.option(
    '--force', '-f',
    is_flag=True,
    help='Ignore any conflicts.'
)
@click.pass_obj
@handle_passthesalt_errors
def pts_pull(pts, path, force):
    """
    Retrieve a remote store.
    """
    pts_path = getattr(pts, 'path', pts)
    remote = read_or_init_remote(path)
    remote_pts = remote.get().with_path(pts_path)
    remote.to_path(path)

    if not isinstance(pts, PassTheSalt) or force or remote_pts.modified > pts.modified:
        remote_pts.save()
        echo('Successfully pulled remote store')
    elif remote_pts == pts:
        echo('Already up to date')
    else:
        bail('local version is newer than the remote')


@cli.command('diff')
@click.option(
    '--path', '-p',
    type=click.Path(),
    default=DEFAULT_REMOTE_PATH,
    help='The path to a local PassTheSalt store or a remote configuration.'
)
@click.option(
    '--kind', '-k',
    type=click.Choice(['encrypted', 'generatable']),
    help='Filter by type of secret.'
)
@click.option(
    '--verbose', '-v',
    count=True,
    help='Increase amount information displayed.'
)
@click.pass_obj
@handle_passthesalt_errors
def pts_diff(pts, path, kind, verbose):
    """
    Compare two stores.
    """
    try:
        remote = read_or_init_remote(path)
        other_pts = remote.get().with_path(pts.path)
        remote.to_path(path)
    except serde.exceptions.DeserializationError:
        other_pts = PassTheSalt.from_path(path)

    diff_left = pts._diff(other_pts)
    diff_right = other_pts._diff(pts)

    if diff_left.labels():
        echo('Local store has the following extra/modified secrets:')
        pts_ls_(diff_left, kind=kind, header=False, verbose=verbose)

    if diff_right.labels():
        if diff_left.labels():
            echo()

        echo('Remote store has the following extra/modified secrets:')
        pts_ls_(diff_right, kind=kind, header=False, verbose=verbose)

    if diff_left.labels() or diff_right.labels():
        exit(1)


@cli.command('migrate', hidden=True)
@click.option(
    '-i', '--input-file',
    type=click.Path(exists=True, dir_okay=False),
    help='The input data file.'
)
@click.pass_obj
@handle_passthesalt_errors
def pts_migrate(pts, input_file):
    """
    Load secrets from a PassTheSalt 2.3.x dumped database.
    """
    if input_file:
        with open(input_file, 'r') as f:
            raw = json.load(f)
    else:
        raw = json.loads(sys.stdin.read())

    for label, raw_secret in raw.items():
        modified = datetime.datetime.strptime(raw_secret['modified'], '%Y%m%d')

        if raw_secret['type'] == 'generatable':
            legacy_algorithm = Algorithm(version=0)

            try:
                domain, username, iteration = raw_secret['salt'].split('|')
                secret = Login(
                    modified=modified,
                    domain=domain,
                    username=username,
                    iteration=int(iteration),
                    algorithm=legacy_algorithm
                )
            except Exception:
                secret = Generatable(
                    modified=modified,
                    salt=raw_secret['salt'],
                    algorithm=legacy_algorithm
                )

            pts.add(label, secret)

        elif raw_secret['type'] == 'encrypted':
            secret = Encrypted(
                raw_secret['secret'],
                modified=modified
            )
            pts.add(label, secret)

        else:
            bail(f'unknown secret type {raw_secret["type"]}')

        echo(f'Migrated {label!r}')

    pts.save()
    echo('Success!')



"""
Remote configuration for PassTheSalt.
"""

import json

import requests
from serde import fields

from passthesalt.core import PassTheSalt
from passthesalt.exceptions import (
    ConflictingTimestamps, RemoteError, UnauthorizedAccess, UnexpectedStatusCode
)
from passthesalt.model import ModifiedModel


class Remote(ModifiedModel):
    """
    Configuration for a remote store.
    """

    location = fields.Url()

    def with_auth(self, auth):
        """
        Configure the Remote with authentication.

        Args:
            auth: the authentication for communicating with the remote store.
                This can be a callback for getting the authentication (for
                example through user input), or the actual authentication.
        """
        self._auth = auth
        return self

    @property
    def auth(self):
        """
        The configured authentication.

        Returns:
            the configured authentication.
        """
        if callable(self._auth):
            self._auth = self._auth(self)

        return self._auth

    @property
    def headers(self):
        """
        General headers to use when making requests to the remote server.

        Returns:
            dict: the headers.
        """
        return {}

    def validate_response(self, response):
        """
        Validate the given response.

        Args:
            response (Response): the requests response.

        Raises:
            UnexpectedStatusCode: for a bad response code.
        """
        try:
            response.raise_for_status()
        except requests.exceptions.HTTPError as e:
            raise UnexpectedStatusCode(str(e) or repr(e), response.status_code)

    def request(self, verb, url, headers=None, auth=None, data=None):
        """
        Make a remote request to the given URL.

        Args:
            verb (str): the HTTP method like 'GET' or 'POST'.
            url (str): the URL to request.
            headers (dict): extra headers to use.
            auth (tuple): authentication to use.
            data (str): the payload.

        Raises:
            RemoteError: when a generic requests exception occurs.

        Returns:
            dict: the server response.
        """
        try:
            response = requests.request(verb, url, headers=headers, auth=auth, data=data)
        except requests.exceptions.RequestException as e:
            raise RemoteError(str(e) or repr(e))
        else:
            self.validate_response(response)
            return response

    def get(self):
        """
        Retrieve the Remote store.

        Returns:
            PassTheSalt: a PassTheSalt instance.
        """
        data = self.request('GET', self.location, headers=self.headers, auth=self.auth)
        return PassTheSalt.from_base64(data)

    def put(self, pts):
        """
        Upload the given PassTheSalt to the Remote store.

        Args:
            pts (PassTheSalt): a PassTheSalt instance.

        Returns:
            Response: the response from the server.
        """
        data = pts.to_base64()
        return self.request('PUT', self.location, headers=self.headers, auth=self.auth, data=data)


class Stow(Remote):
    """
    Stow configuration for a remote store.

    See https://github.com/rossmacarthur/stow.
    """

    token = fields.Optional(fields.Str)
    token_location = fields.Url()

    @property
    def headers(self):
        """
        General headers to use when making requests to the remote server.

        Returns:
            dict: the headers.
        """
        return {
            'Content-Type': 'application/json'
        }

    def validate_response(self, response):
        """
        Validate the given response.

        Args:
            response (Response): the requests response.

        Raises:
            UnauthorizedAccess: when the response code is 401.
            ConflictingTimestamps: when the response code is 409.
        """
        try:
            message = response.json().get('message')
        except json.decoder.JSONDecodeError:
            message = None

        if response.status_code == 401:
            raise UnauthorizedAccess(message)
        elif response.status_code == 409:
            raise ConflictingTimestamps(message)

        super().validate_response(response)

    def renew(self):
        """
        Renew the internal token with the configured authentication.
        """
        data = self.request(
            'GET', self.token_location, headers=self.headers, auth=self.auth
        ).json()
        self.token = data['token']
        self.touch()

    def handle_renew(self, f):
        """
        Decorator to recall a function if the token needs to be renewed.

        If the given function raises UnauthorizedAccess then renew the token and
        recall the function.

        Args:
            f (callable): the function that accesses some secure resource.

        Returns:
            the result of the function.
        """
        def decorated_function(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except UnauthorizedAccess:
                self.renew()
                return f(*args, **kwargs)

        return decorated_function

    def get(self):
        """
        Retrieve the Remote store.

        Returns:
            PassTheSalt: a PassTheSalt instance.
        """
        @self.handle_renew
        def get():
            return self.request(
                'GET', self.location, headers=self.headers, auth=(self.token, 'unused')
            ).json()

        return PassTheSalt.from_base64(get()['value'])

    def put(self, pts, force=False):
        """
        Upload the given PassTheSalt to the Remote store.

        Args:
            pts (PassTheSalt): a PassTheSalt instance.
            force (bool): whether to ignore any conflicts.

        Returns:
            dict: the message from the server.
        """
        @self.handle_renew
        def put(pts):
            payload = {
                'value': pts.to_base64()
            }

            if not force:
                payload['modified'] = pts.modified.isoformat()

            data = json.dumps(payload)
            return self.request(
                'PUT', self.location, headers=self.headers, auth=(self.token, 'unused'), data=data
            )

        try:
            message = put(pts).json().get('message')
        except json.decoder.JSONDecodeError:
            message = None

        return message



"""
Errors used in PassTheSalt.
"""


class PassTheSaltError(Exception):
    """
    A base error class.
    """

    def __init__(self, message):
        """
        Create a new PassTheSaltError.

        Args:
            message (str): the error message.
        """
        super().__init__(message)

    @property
    def message(self):
        """
        Return the error message.
        """
        return self.args[0]

    def __str__(self):
        """
        Return a string representation of this PassTheSaltError.
        """
        return self.message

    def __repr__(self):
        """
        Return the canonical string representation of this PassTheSaltError.
        """
        return f'{self.__class__.__module__}.{self.__class__.__name__}({self.message!r})'


class LabelError(PassTheSaltError):
    """
    An error related to secret labels.
    """


class ContextError(PassTheSaltError):
    """
    An error related to the PassTheSalt context of a secret.
    """


class ConfigurationError(PassTheSaltError):
    """
    An error related to the PassTheSalt configuration.
    """


class RemoteError(PassTheSaltError):
    """
    An error related to accessing a remote store.
    """


class UnexpectedStatusCode(RemoteError):
    """
    Raised when an unexpected HTTP status code is received.
    """

    def __init__(self, message, code):
        """
        Create a new UnexpectedStatusCode error.

        Args:
            message (str): the error message.
            code (int): the response status code.
        """
        super().__init__(message)
        self.code = code

    def __repr__(self):
        """
        Return the canonical string representation of this SerdeError.
        """
        return (
            f'{self.__class__.__module__}.{self.__class__.__name__}'
            f'({self.message!r}, code={self.code!r})'
        )


class UnauthorizedAccess(UnexpectedStatusCode):
    """
    Raised when a 401 is received.
    """

    def __init__(self, message, code=401):
        """
        Create a new UnauthorizedAccess error.

        Args:
            message (str): the error message.
            code (int): the response status code.
        """
        super().__init__(message, code)


class ConflictingTimestamps(UnexpectedStatusCode):
    """
    Raised when a 409 is received.
    """

    def __init__(self, message, code=409):
        """
        Create a new ConflictingTimestamps error.

        Args:
            message (str): the error message.
            code (int): the response status code.
        """
        super().__init__(message, code)



"""
PassTheSalt is a deterministic password generation and password storage system.
"""

from passthesalt.core import (
    Algorithm, Config, Encrypted, Generatable, Login, Master, PassTheSalt, Secret
)


__all__ = [
    'Algorithm',
    'Config',
    'Encrypted',
    'Generatable',
    'Login',
    'Master',
    'PassTheSalt',
    'Secret',
    'error',
    'remote'
]
__title__ = 'passthesalt'
__version__ = '3.1.0'
__url__ = 'https://github.com/rossmacarthur/passthesalt'
__author__ = 'Ross MacArthur'
__author_email__ = 'ross@macarthur.io'
__license__ = 'MIT'
__description__ = 'Deterministic password generation and password storage.'



from binascii import hexlify

from passthesalt.crypto import (
    decrypt, encrypt, generate, generate_key, passlify,
    passlify_legacy, pbkdf2_hash, pbkdf2_hash_bytes
)


def test_pbkdf2_hash_bytes():
    expected = (
        b'#\x1a\xfb}\xcd.\x86\x0c\xfdX\xab\x137+\xd1,'
        b'\x920v\xc3Y\x8a\x12\x19`2\x0fo\xec\x8aV\x98'
    )
    assert pbkdf2_hash_bytes('password', b'salt', iterations=1024, length=32) == expected


def test_pbkdf2_hash_own_salt():
    salt = hexlify(b'salt')
    expected = (salt, '231afb7dcd2e860cfd58ab13372bd12c923076c3')
    assert pbkdf2_hash('password', salt=salt, iterations=1024) == expected


def test_pbkdf2_hash_gen_salt():
    password = 'password'
    salt, hash = pbkdf2_hash(password)
    assert pbkdf2_hash(password, salt=salt)[1] == hash


def test_generate_key():
    assert generate_key('password') == b'XohImNooBHFR0OVvjcYpJ3NgPQ1qq73WKhHvch0VQtg='


def test_encrypt_and_decrypt():
    master_key = 'password'

    for d in ({}, {'b': 'herp', 'a': 'derp'}, {'a': 5}):
        encrypted = encrypt(d, master_key)
        assert decrypt(encrypted, master_key) == d


def test_passlify_legacy():
    b = b'\x00'
    assert passlify_legacy(b) == 'b'

    b = b'\x35\x00\x26\x48'
    assert passlify_legacy(b) == 'aM-1'


def test_passlify():
    b = b'\x00'
    assert passlify(b) == 'B'

    b = b'\x00\x00\x00\x00'
    assert passlify(b) == 'Ty_7'


def test_generate():
    # these values are taken from PassTheSalt v2.2.1
    assert generate('salt', 'John Smith|password') == 'wDG4-5D$aexGL+gE*8HQ'
    assert generate('salt2', 'John Smith|password') == 'oitKf=l2ziz9*6MH%e@$'
    assert generate('salt', 'John Smith|password', version=1) == 'wDG4-5D$aexGL+gE*8HQ'
    assert generate('salt2', 'John Smith|password', version=1) == 'oitKf=l2ziz9*6MH%e@$'



import datetime
import tempfile

from pytest import raises

from passthesalt.model import DateTime, ModifiedModel


class TestDateTime:

    def test___init__(self):
        example = DateTime(validators=[])
        assert example.validators == []
        assert not hasattr(example, 'format')

    def test_serialize(self):
        value = datetime.datetime(year=2017, month=12, day=29)
        assert DateTime().serialize(value) == '2017-12-29'

        value = datetime.datetime(year=2017, month=12, day=29, hour=10)
        assert DateTime().serialize(value) == '2017-12-29 10:00:00'

        value = datetime.datetime(year=2017, month=12, day=29, microsecond=1000)
        assert DateTime().serialize(value) == '2017-12-29 00:00:00.001000'

    def test_deserialize(self):
        expected = datetime.datetime(year=2017, month=12, day=29)
        assert DateTime().deserialize('2017-12-29') == expected
        assert DateTime().deserialize('2017-12-29 00:00:00') == expected
        assert DateTime().deserialize('2017-12-29 00:00:00.000000') == expected

    def test_deserialize_invalid(self):
        with raises(ValueError):
            DateTime().deserialize('2017-12-29T00:00:00')


class TestModifiedModel:

    def test_from_base64(self):
        value = 'eyJtb2RpZmllZCI6ICIyMDE4LTEyLTI1IDAwOjAwOjAwIn0='
        expected = ModifiedModel(
            modified=datetime.datetime(year=2018, month=12, day=25)
        )
        assert ModifiedModel.from_base64(value) == expected

    def test_from_path(self):
        example = ModifiedModel(
            modified=datetime.datetime(year=2018, month=12, day=25)
        )
        with tempfile.NamedTemporaryFile() as t:
            with open(t.name, 'w') as f:
                f.write(example.to_json())
            assert ModifiedModel.from_path(t.name) == example

    def test_to_base64(self):
        example = ModifiedModel(
            modified=datetime.datetime(year=2018, month=12, day=25)
        )
        expected = 'eyJtb2RpZmllZCI6ICIyMDE4LTEyLTI1In0='
        assert example.to_base64() == expected

    def test_to_path(self):
        example = ModifiedModel()
        with tempfile.NamedTemporaryFile() as t:
            example.to_path(t.name, sort_keys=True)
            with open(t.name) as f:
                assert f.read() == example.to_json(sort_keys=True)



import datetime
import string
import tempfile
from unittest import mock

import serde
from pytest import raises

from passthesalt import __version__
from passthesalt.core import (
    Config, Encrypted, Generatable, Login, Master, PassTheSalt, Secret, version
)
from passthesalt.exceptions import ConfigurationError, ContextError, LabelError


class SecretSub(Secret):
    pass


def test_version():
    assert version() == __version__


@mock.patch.dict('passthesalt.core.SECRETS', {'secret': SecretSub})
@mock.patch.dict('passthesalt.core.KINDS', {SecretSub: 'secret'})
class TestSecret:

    def test_from_dict(self):
        given = {
            'kind': 'secret',
            'modified': '2017-12-29'
        }
        expected = SecretSub(
            modified=datetime.datetime(year=2017, month=12, day=29)
        )
        assert Secret.from_dict(given) == expected

    def test_from_dict_invalid_kind(self):
        with raises(serde.exceptions.DeserializationError):
            Secret.from_dict({
                'kind': 'unknown',
                'modified': '2017-12-29'
            })

    def test_to_dict(self):
        given = SecretSub(
            modified=datetime.datetime(year=2017, month=12, day=29)
        )
        expected = {
            'kind': 'secret',
            'modified': '2017-12-29'
        }
        assert given.to_dict() == expected

    def test_to_dict_unknown_kind(self):
        class SecretSub2(Secret):
            pass

        with raises(serde.exceptions.SerializationError):
            SecretSub2().to_dict()

    def test___getattr__(self):
        example = Secret()

        with raises(ContextError):
            example._pts

        with raises(ContextError):
            example._label

        with raises(AttributeError):
            example._something

    def test_display(self):
        example = SecretSub()
        example.add_context('example', object())
        assert example.display() == ('example', 'secret', example.modified)

    def test_add_context(self):
        example = SecretSub()
        label = 'example'
        pts = object()
        example.add_context(label, pts)
        assert example._label == 'example'
        assert example._pts == pts

    def test_check_context(self):
        example = SecretSub()

        with raises(ContextError):
            example.check_context()

        example.add_context('example', object())

        assert example.check_context() is None

    def test_remove_context(self):
        example = SecretSub()
        example.add_context('example', object())
        example.remove_context()

        with raises(ContextError):
            example.check_context()

    def test_get(self):
        with raises(NotImplementedError):
            Secret().get()


class TestGeneratable:

    def test_display(self):
        modified = datetime.datetime(year=2018, month=12, day=25)
        secret = Generatable(salt='test', modified=modified)
        secret._label = 'Example'
        assert secret.display() == ('Example', 'generatable', modified, 'test')

    def test_get(self):
        pts = PassTheSalt().with_master('password')
        secret = Generatable(salt='test')
        secret.add_context('Example', pts)
        assert secret.get() == 'CV*2qua!A3rwh0fwf8o*'


class TestLogin:

    def test_salt(self):
        pts = PassTheSalt().with_master('password')
        secret = Login(domain='www.test.com', username='test')
        secret.add_context('Example', pts)
        assert secret.salt == 'www.test.com|test|0'
        assert secret.get() == 'j4RETtUP7xyQcR%Lc1k+'


class TestEncrypted:

    def test___init__(self):
        secret = Encrypted('verysecure')
        assert secret.secret == 'verysecure'

    def test__encrypt_and__decrypt(self):
        pts = PassTheSalt().with_master('password')
        secret = Encrypted('verysecure')
        secret.add_context('Example', pts)
        assert pts.secrets_encrypted is None
        assert secret._decrypt() == {}
        secret._encrypt({'Example': secret.secret})
        assert secret._decrypt() == {'Example': secret.secret}

    def test_add(self):
        pts = PassTheSalt().with_master('password')
        secret = Encrypted('verysecure')
        secret.add_context('Example', pts)
        assert pts.secrets_encrypted is None
        secret.add()
        assert bool(pts.secrets_encrypted)
        assert not hasattr(secret, 'secret')

    def test_get(self):
        pts = PassTheSalt().with_master('password')
        secret = Encrypted('verysecure')
        secret.add_context('Example', pts)
        secret.add()
        assert secret.get() == 'verysecure'

    def test_get_missing(self):
        pts = PassTheSalt().with_master('password')
        secret = Encrypted('verysecure')
        secret.add_context('Example', pts)
        secret.add()
        pts.secrets_encrypted = None
        with raises(LabelError):
            secret.get()

    def test_remove(self):
        pts = PassTheSalt().with_master('password')
        secret = Encrypted('verysecure')
        secret.add_context('Example', pts)
        secret.add()
        secret.remove()
        assert pts.secrets_encrypted is None


class TestMaster:

    def test___init__(self):
        def is_hexstring(s):
            return (
                isinstance(s, str)
                and len(s) == 40
                and all(h in string.hexdigits for h in s)
            )

        example = Master('password')
        assert is_hexstring(example.hash)
        assert is_hexstring(example.salt)
        assert not hasattr(example, 'password')

    def test_is_valid(self):
        master = Master('password')
        assert master.is_valid('password') is True
        assert master.is_valid('password2') is False


class TestPassTheSalt:

    def test_from_dict_no_secrets(self):
        given = {
            'config': {
                'master': {
                    'salt': 'b32ad5ad536b2e3b790050845b65311b333e6480',
                    'hash': '300148d88175b2d9cf52001f29e716de1b3a56ea'
                },
                'owner': 'John Smith'
            },
            'modified': '2018-12-25'
        }
        expected = PassTheSalt(
            config=Config(
                owner='John Smith',
                master=Master.from_dict({
                    'salt': 'b32ad5ad536b2e3b790050845b65311b333e6480',
                    'hash': '300148d88175b2d9cf52001f29e716de1b3a56ea'
                })
            ),
            modified=datetime.datetime(year=2018, month=12, day=25)
        )
        assert PassTheSalt.from_dict(given) == expected

    def test_from_dict_secrets(self):
        given = {
            'config': {
                'master': {
                    'salt': 'b32ad5ad536b2e3b790050845b65311b333e6480',
                    'hash': '300148d88175b2d9cf52001f29e716de1b3a56ea'
                },
                'owner': 'John Smith'
            },
            'secrets': {
                'Example': {
                    'algorithm': {
                        'version': 1
                    },
                    'salt': 'test',
                    'kind': 'generatable',
                    'modified': '2018-12-25'
                }
            },
            'modified': '2018-12-25'
        }
        expected = PassTheSalt(
            config=Config(
                owner='John Smith',
                master=Master.from_dict({
                    'salt': 'b32ad5ad536b2e3b790050845b65311b333e6480',
                    'hash': '300148d88175b2d9cf52001f29e716de1b3a56ea'
                })
            )
        )
        modified = datetime.datetime(year=2018, month=12, day=25)
        expected.add('Example', Generatable(modified=modified, salt='test'))
        expected.modified = modified
        assert PassTheSalt.from_dict(given) == expected

    def test_save(self):
        with tempfile.NamedTemporaryFile() as t:
            pts = PassTheSalt().with_path(t.name)
            pts.save()
            assert PassTheSalt.from_path(t.name) == pts

    def test_with_master(self):
        pts = PassTheSalt()
        assert pts.with_master('password') is pts
        assert pts._master == 'password'

    def test_with_path(self):
        pts = PassTheSalt()
        assert pts.with_path('example') is pts
        assert pts._path == 'example'

    def test_master_key_empty(self):
        pts = PassTheSalt()
        with raises(ConfigurationError):
            pts.master_key

    def test_master_key_no_owner(self):
        pts = PassTheSalt()
        assert pts.with_master('password').master_key == 'password'

    def test_master_key_owner(self):
        pts = PassTheSalt(config=Config(owner='John Smith'))
        assert pts.with_master('password').master_key == 'John Smith|password'

    def test_master_key_callable_master(self):
        pts = PassTheSalt()
        assert pts.with_master(lambda x: 'password').master_key == 'password'

    def test_path(self):
        pts = PassTheSalt().with_path('test')
        assert pts.path == 'test'

    def test_path_empty(self):
        pts = PassTheSalt()
        with raises(ConfigurationError):
            pts.path

    def test_labels_no_pattern(self):
        pts = PassTheSalt()
        pts.add('Example', Generatable(salt='test'))
        assert pts.labels() == ['Example']

    def test_labels_good_pattern(self):
        pts = PassTheSalt()
        pts.add('Example', Generatable(salt='test'))
        pts.add('Example2', Generatable(salt='test'))
        pts.add('3Example', Generatable(salt='test'))
        expected = {'Example', 'Example2', '3Example'}
        assert set(pts.labels(pattern=r'\d?Example\d?')) == expected

    def test_labels_bad_pattern(self):
        pts = PassTheSalt()
        with raises(LabelError):
            pts.labels(pattern='(')

    def test_resolve_not_exists(self):
        pts = PassTheSalt()
        with raises(LabelError):
            pts.resolve(pattern='Example')

    def test_resolve_exact(self):
        pts = PassTheSalt()
        pts.add('Example', Generatable(salt='test'))
        assert pts.resolve(pattern='Example') == 'Example'

    def test_resolve_exists(self):
        pts = PassTheSalt()
        pts.add('Example', Generatable(salt='test'))
        assert pts.resolve(pattern='[aelmpxE]{7}') == 'Example'

    def test_resolve_multiple(self):
        pts = PassTheSalt()
        pts.add('Example', Generatable(salt='test'))
        pts.add('Example2', Generatable(salt='test'))
        with raises(LabelError):
            pts.resolve(pattern=r'^Example\d?$')

    def test_resolve_one(self):
        pts = PassTheSalt()
        pts.add('Example', Generatable(salt='test'))
        pts.add('Example2', Generatable(salt='test'))
        assert pts.resolve(pattern=r'^Example$') == 'Example'

    def test_contains(self):
        pts = PassTheSalt()
        pts.add('Example', Generatable(salt='test'))
        assert pts.contains('Example') is True
        assert pts.contains('Example2') is False

    def test_add_not_exists(self):
        pts = PassTheSalt()
        secret = Generatable(salt='salt')
        with raises(ContextError):
            secret.check_context()
        pts.add('Example', secret)
        assert secret._pts is pts
        assert secret._label == 'Example'

    def test_add_exists(self):
        pts = PassTheSalt()
        pts.add('Example', Generatable(salt='salt'))
        with raises(LabelError):
            pts.add('Example', Generatable(salt='salt'))

    def test_get(self):
        pts = PassTheSalt()
        secret = Generatable(salt='salt')
        pts.add('Example', secret)
        assert pts.get('Example') is secret

    def test_pop(self):
        pts = PassTheSalt()
        secret = Generatable(salt='salt')
        pts.add('Example', secret)
        assert pts.pop('Example') is secret
        assert not pts.contains('Example')
        with raises(ContextError):
            secret._pts

    def test_pop_not_exists(self):
        pts = PassTheSalt()
        with raises(LabelError):
            pts.pop('Example')

    def test_remove(self):
        pts = PassTheSalt()
        secret = Generatable(salt='salt')
        pts.add('Example', secret)
        assert pts.contains('Example')
        pts.remove('Example')
        assert not pts.contains('Example')
        with raises(ContextError):
            secret._pts
        with raises(ContextError):
            secret._label

    def test_move(self):
        pts = PassTheSalt()
        pts.add('Example', Generatable(salt='salt'))
        assert pts.contains('Example')
        assert not pts.contains('Example2')
        pts.move('Example', 'Example2')
        assert not pts.contains('Example')
        assert pts.contains('Example2')

    def test_move_already_exists(self):
        pts = PassTheSalt()
        pts.add('Example', Generatable(salt='salt'))
        with raises(LabelError):
            pts.move('test', 'Example')

    def test_update_not_exists(self):
        pts = PassTheSalt()
        secret = Generatable(salt='salt')
        with raises(ContextError):
            secret.check_context()
        pts.update('Example', secret)
        assert secret._pts is pts
        assert secret._label == 'Example'

    def test_update_already_exists(self):
        pts = PassTheSalt()
        secret = Generatable(salt='salt')
        secret2 = Generatable(salt='test')
        pts.add('Example', secret)
        pts.update('Example', secret2)
        assert pts.get('Example') is secret2
        with raises(ContextError):
            secret._pts
        with raises(ContextError):
            secret._label

    def test__diff(self):
        pts = PassTheSalt()
        other = PassTheSalt()
        secret = Generatable(salt='salt')
        pts.add('Example', secret)
        diff = pts._diff(other)
        assert diff.get('Example') == secret
        assert not other._diff(pts).labels()



"""
Cryptographic utilities used in PassTheSalt.
"""

import json
import os
import string
from base64 import urlsafe_b64encode
from binascii import hexlify, unhexlify
from hashlib import sha256

from cryptography.fernet import Fernet
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC


def pbkdf2_hash_bytes(password, salt, iterations=2048, length=20):
    """
    PBKDF2 HMAC-SHA-256 hash a password.

    Args:
        password (str): the password to hash.
        salt (bytes): the salt for the hash.
        iterations (int): the number of PBKDF2 iterations.

    Returns:
        bytes: the hash.
    """
    return PBKDF2HMAC(
        algorithm=hashes.SHA256(),
        length=length,
        salt=salt,
        iterations=iterations,
        backend=default_backend()
    ).derive(password.encode())


def pbkdf2_hash(password, salt=None, iterations=100000):
    """
    PBKDF2 HMAC-SHA-256 hash a password.

    Args:
        password (str): the password to hash.
        salt (str): the salt for the hash.
        iterations (int): the number of PBKDF2 iterations.

    Returns:
        str: the hash as a hex string.
    """
    if salt is None:
        salt = hexlify(os.urandom(20)).decode()

    hash_bytes = pbkdf2_hash_bytes(password, unhexlify(salt), iterations=iterations)
    hash = hexlify(hash_bytes).decode()

    return salt, hash


def generate_key(master_key):
    """
    Generate a Fernet key from the master key.

    Args:
        master_key (str): the master key.

    Returns:
        str: 32 url-safe base64-encoded bytes.
    """
    return urlsafe_b64encode(sha256(master_key.encode()).digest())


def encrypt(d, master_key):
    """
    Encrypt a dictionary with a master key.

    Args:
        d (dict): the dictionary to encrypt.
        master_key (str): the master key.

    Returns:
        str: the encrypted dictionary as a hex string.
    """
    key = generate_key(master_key)
    fernet = Fernet(key)
    data = json.dumps(d, sort_keys=True).encode()
    encrypted = fernet.encrypt(data)
    return hexlify(encrypted).decode()


def decrypt(s, master_key):
    """
    Decrypt an encrypted dictionary with the master key.

    Args:
        s (str): the encrypted hexstring.
        master_key (str): the master key.

    Returns:
        dict: the decrypted dictionary.
    """
    key = generate_key(master_key)
    fernet = Fernet(key)
    decrypted = fernet.decrypt(unhexlify(s))
    return json.loads(decrypted.decode())


def passlify_legacy(b):
    """
    Create a legacy password from a bytes object.

    WARNING: This method is broken. It is kept so that legacy passwords from
    older versions of PassTheSalt can still be generated.

    Args:
        b (bytes): the bytes object.

    Returns:
        str: the password.
    """
    def shiftlify(b, chars, shift=0):
        b = [(byte + shift) % 256 for byte in b]
        return ''.join(chars[byte % len(chars)] for byte in b)

    groups = [string.ascii_lowercase, string.ascii_uppercase, string.digits, '!@#$%^&*_+-=']
    chars = ''.join(groups)
    password = shiftlify(b, chars)

    # This make sure the password contains something in each group
    for _ in range(256):
        if all(any(c in group for c in password) for group in groups):
            break

        password = shiftlify(b, chars, shift=1)

    # Attempt to make sure the password starts with a letter
    for _ in range(len(password)):
        if password[0] in string.ascii_letters:
            break

        password = password[1:] + password[0:1]

    return password


def passlify(b, lowers=True, uppers=True, digits=True, extras='!@#$%^&*_+-=', startswith=None):
    """
    Create a password from a bytes object.

    This method attempts to make sure the password contains at least one
    lowercase letter, uppercase letter, a number and a symbol. It also shifts
    the password until the first character is a letter.

    Args:
        b (bytes): the bytes object.
        lowers (bool): use lowercase letters.
        uppers (bool): use uppercase letters.
        digits (bool): use digits.
        extras (str): extra symbols to use.
        startswith (str): attempt to make sure the password starts with one of
            these chracters.

    Returns:
        str: the password.
    """
    groups = []

    if lowers:
        groups.append(string.ascii_lowercase)
    if uppers:
        groups.append(string.ascii_uppercase)
    if digits:
        groups.append(string.digits)
    if extras:
        groups.append(extras)
    if startswith is None:
        startswith = string.ascii_letters

    chars = ''.join(groups)

    def shift(i):
        return 7 * ((i + 1)**2 + (i + 1)) / 2

    def rotate(b):
        b = [int(byte + shift(i)) % 256 for i, byte in enumerate(b)]
        return bytes(b[1:] + b[0:1])

    # This make sure the password contains something in each group
    for _ in range(256):
        password = ''.join(chars[byte % len(chars)] for byte in b)

        if all(any(c in group for c in password) for group in groups):
            break

        b = rotate(b)

    # Attempt to make sure the password starts with a letter
    for _ in range(len(password)):
        if password[0] in startswith:
            break

        password = password[1:] + password[0:1]

    return password


def generate(salt, master_key, version=None, length=None):
    """
    Generate a password from a salt and master key.

    Args:
        salt (str): the salt.
        master_key (str): the master key.
        version (int): the generation algorithm version.

    Returns:
        str: the generated secret.
    """
    if length is None:
        length = 20

    b = pbkdf2_hash_bytes(master_key, salt=salt.encode(), length=length)

    if version == 1:
        return passlify(b)

    return passlify_legacy(b)



from unittest import mock

import click
from click.testing import CliRunner
from pytest import raises

from passthesalt import Config, Encrypted, Generatable, Login, Master, PassTheSalt
from passthesalt.cli import bail, cli, handle_passthesalt_errors
from passthesalt.exceptions import PassTheSaltError


def test_bail():
    with raises(click.ClickException) as e:
        bail('test')
        assert e.message == 'test'


def test_handle_passthesalt_errors():

    @handle_passthesalt_errors
    def raise_passthesalt_error():
        raise PassTheSaltError('test')

    with raises(click.ClickException) as e:
        raise_passthesalt_error()
        assert e.message == 'test'


def test_cli_initializing():
    runner = CliRunner()

    with runner.isolated_filesystem():
        result = runner.invoke(
            cli,
            ['--path', 'passthesalt'],
            input='John Smith\npassword\npassword\n'
        )
        assert result.exit_code == 0
        assert 'Initializing PassTheSalt ...' in result.output
        assert 'Successfully initialized PassTheSalt!' in result.output

        pts = PassTheSalt.from_path('passthesalt')
        assert pts.config.owner == 'John Smith'
        assert pts.config.master
        assert pts.config.master.is_valid('password')


def test_cli_missing_command():
    runner = CliRunner()

    with runner.isolated_filesystem():
        PassTheSalt().to_path('passthesalt')

        result = runner.invoke(
            cli,
            ['--path', 'passthesalt']
        )
        assert result.exit_code == 0
        assert 'Usage: cli [OPTIONS] COMMAND [ARGS]...' in result.output


def test_pts_add_raw():
    runner = CliRunner()

    with runner.isolated_filesystem():
        PassTheSalt().to_path('passthesalt')
        result = runner.invoke(
            cli,
            ['--path', 'passthesalt', 'add', '--type', 'raw'],
            input='Example\nsalt\ny\nn\n'
        )
        assert result.exit_code == 0
        assert "Stored 'Example'!" in result.output

        pts = PassTheSalt.from_path('passthesalt')
        assert isinstance(pts.get('Example'), Generatable)
        assert pts.get('Example').salt == 'salt'


def test_pts_add_login():
    runner = CliRunner()

    with runner.isolated_filesystem():
        PassTheSalt().to_path('passthesalt')
        result = runner.invoke(
            cli,
            ['--path', 'passthesalt', 'add'],
            input='Example\nwww\nwww.test.com\ntest\n\ny\nn\n'
        )
        assert result.exit_code == 0
        assert "Stored 'Example'!" in result.output

        pts = PassTheSalt.from_path('passthesalt')
        assert isinstance(pts.get('Example'), Login)
        assert pts.get('Example').salt == 'www.test.com|test|0'


def test_pts_add_exists():
    runner = CliRunner()

    with runner.isolated_filesystem():
        pts = PassTheSalt()
        pts.add('Example', Generatable(salt='salt'))
        pts.to_path('passthesalt')

        result = runner.invoke(
            cli,
            ['--path', 'passthesalt', 'add'],
            input='Example\n'
        )
        assert result.exit_code == 1
        assert "'Example' already exists" in result.output


def test_pts_encrypt():
    runner = CliRunner()

    with runner.isolated_filesystem():
        pts = PassTheSalt(config=Config(master=Master('password')))
        pts.to_path('passthesalt')

        result = runner.invoke(
            cli,
            ['--path', 'passthesalt', 'encrypt'],
            input='Example\nsecret\nsecret\npassword\n'
        )
        assert result.exit_code == 0
        assert "Stored 'Example'!" in result.output

        pts = PassTheSalt.from_path('passthesalt').with_master('password')
        assert isinstance(pts.get('Example'), Encrypted)
        assert pts.get('Example').get() == 'secret'


def test_pts_encrypt_exists():
    runner = CliRunner()

    with runner.isolated_filesystem():
        pts = PassTheSalt()
        pts.add('Example', Generatable(salt='salt'))
        pts.to_path('passthesalt')

        result = runner.invoke(
            cli,
            ['--path', 'passthesalt', 'encrypt'],
            input='Example\n'
        )
        assert result.exit_code == 1
        assert "'Example' already exists" in result.output


def pts_edit_generatable_setup():
    pts = PassTheSalt()
    pts.add('Example', Generatable(salt='salt'))
    pts.to_path('passthesalt')
    return pts


def test_pts_edit_generatable():
    def edit(data):
        return data.replace('salt = "salt"', 'salt = "test"')

    with mock.patch('click.edit', edit):
        runner = CliRunner()

        with runner.isolated_filesystem():
            pts = pts_edit_generatable_setup()
            result = runner.invoke(
                cli,
                ['--path', 'passthesalt', 'edit'],
                input='Example\nn\n'
            )
            assert result.exit_code == 0
            assert "Updated 'Example'!\n" in result.output

            pts = PassTheSalt.from_path('passthesalt')
            assert isinstance(pts.get('Example'), Generatable)
            assert pts.get('Example').salt == 'test'


def test_pts_edit_generatable_unchanged():
    def edit(data):
        return data

    with mock.patch('click.edit', edit):
        runner = CliRunner()

        with runner.isolated_filesystem():
            pts = pts_edit_generatable_setup()
            result = runner.invoke(
                cli,
                ['--path', 'passthesalt', 'edit'],
                input='Example\nn\n'
            )
            assert result.exit_code == 0
            assert result.output.endswith("'Example' was not changed\n")

            pts = PassTheSalt.from_path('passthesalt')
            assert isinstance(pts.get('Example'), Generatable)
            assert pts.get('Example').salt == 'salt'


def test_pts_edit_generatable_aborted():
    def edit(data):
        return None

    with mock.patch('click.edit', edit):
        runner = CliRunner()

        with runner.isolated_filesystem():
            pts = pts_edit_generatable_setup()
            result = runner.invoke(
                cli,
                ['--path', 'passthesalt', 'edit'],
                input='Example\nn\n'
            )
            assert result.exit_code == 1
            assert result.output.endswith('Aborted!\n')

            pts = PassTheSalt.from_path('passthesalt')
            assert isinstance(pts.get('Example'), Generatable)
            assert pts.get('Example').salt == 'salt'


def pts_edit_encrypted_setup():
    pts = PassTheSalt(config=Config(master=Master('password'))).with_master('password')
    pts.add('Example', Encrypted('verysecret'))
    pts.to_path('passthesalt')
    return pts


def test_pts_edit_encrypted():
    def edit(data):
        return 'notverysecret'

    with mock.patch('click.edit', edit):
        runner = CliRunner()

        with runner.isolated_filesystem():
            pts = pts_edit_encrypted_setup()
            result = runner.invoke(
                cli,
                ['--path', 'passthesalt', 'edit'],
                input='Example\npassword\nn\n'
            )
            assert result.exit_code == 0
            assert "Updated 'Example'!\n" in result.output

            pts = PassTheSalt.from_path('passthesalt').with_master('password')
            assert isinstance(pts.get('Example'), Encrypted)
            assert pts.get('Example').get() == 'notverysecret'


def test_pts_edit_encrypted_unchanged():
    def edit(data):
        return data

    with mock.patch('click.edit', edit):
        runner = CliRunner()

        with runner.isolated_filesystem():
            pts = pts_edit_encrypted_setup()
            result = runner.invoke(
                cli,
                ['--path', 'passthesalt', 'edit'],
                input='Example\npassword\nn\n'
            )
            assert result.exit_code == 0
            assert result.output.endswith("'Example' was not changed\n")

            pts = PassTheSalt.from_path('passthesalt').with_master('password')
            assert isinstance(pts.get('Example'), Encrypted)
            assert pts.get('Example').get() == 'verysecret'


def test_pts_edit_encrypted_aborted():
    def edit(data):
        return None

    with mock.patch('click.edit', edit):
        runner = CliRunner()

        with runner.isolated_filesystem():
            pts = pts_edit_encrypted_setup()
            result = runner.invoke(
                cli,
                ['--path', 'passthesalt', 'edit'],
                input='Example\npassword\nn\n'
            )
            assert result.exit_code == 1
            assert result.output.endswith('Aborted!\n')

            pts = PassTheSalt.from_path('passthesalt').with_master('password')
            assert isinstance(pts.get('Example'), Encrypted)
            assert pts.get('Example').get() == 'verysecret'


def test_pts_get_generated():
    runner = CliRunner()

    with runner.isolated_filesystem():
        pts = PassTheSalt(config=Config(master=Master('password')))
        pts.add('Example', Generatable(salt='salt'))
        pts.to_path('passthesalt')

        result = runner.invoke(
            cli,
            ['--path', 'passthesalt', 'get', '--no-clipboard'],
            input='Example\npassword\n'
        )
        assert result.exit_code == 0
        assert 'M%J+hUIcYqe=LSDtSq0d' in result.output


def test_pts_get_incorrect_master():
    runner = CliRunner()

    with runner.isolated_filesystem():
        pts = PassTheSalt(config=Config(master=Master('password')))
        pts.add('Example', Generatable(salt='salt'))
        pts.to_path('passthesalt')

        result = runner.invoke(
            cli,
            ['--path', 'passthesalt', 'get', '--no-clipboard'],
            input='Example\npasswor\npasswor\npasswor\n'
        )
        assert result.exit_code == 1
        assert 'three incorrect attempts' in result.output


def test_pts_get_encrypted():
    runner = CliRunner()

    with runner.isolated_filesystem():
        pts = PassTheSalt(config=Config(master=Master('password'))).with_master('password')
        pts.add('Example', Encrypted('verysecret'))
        pts.to_path('passthesalt')

        result = runner.invoke(
            cli,
            ['--path', 'passthesalt', 'get', '--no-clipboard'],
            input='Example\npassword\n'
        )
        assert result.exit_code == 0
        assert 'verysecret' in result.output


def test_pts_ls_none():
    runner = CliRunner()

    with runner.isolated_filesystem():
        PassTheSalt().to_path('passthesalt')
        result = runner.invoke(
            cli,
            ['--path', 'passthesalt', 'ls']
        )
        assert 'No stored secrets' in result.output


def test_pts_ls():
    runner = CliRunner()

    with runner.isolated_filesystem():
        pts = PassTheSalt().with_master('password')
        pts.add('Example1', Generatable(salt='salt'))
        pts.add('Example2', Encrypted('verysecret'))
        pts.to_path('passthesalt')
        result = runner.invoke(
            cli,
            ['--path', 'passthesalt', 'ls', '-v']
        )
        expected = (
            'Label     Kind\n'
            '--------  -----------\n'
            'Example1  generatable\n'
            'Example2  encrypted\n'
        )
        assert result.output == expected


def test_pts_rm():
    runner = CliRunner()

    with runner.isolated_filesystem():
        pts = PassTheSalt()
        pts.add('Example', Generatable(salt='salt'))
        pts.to_path('passthesalt')

        result = runner.invoke(
            cli,
            ['--path', 'passthesalt', 'rm'],
            input='Example\ny\n'
        )
        assert result.exit_code == 0
        assert "Removed 'Example'!" in result.output


def test_pts_rm_not_exists():
    runner = CliRunner()

    with runner.isolated_filesystem():
        PassTheSalt().to_path('passthesalt')

        result = runner.invoke(
            cli,
            ['--path', 'passthesalt', 'rm'],
            input='Example\n'
        )
        assert result.exit_code == 1
        assert "'Example' does not exist" in result.output


def test_pts_rm_multiple_not_exists():
    runner = CliRunner()

    with runner.isolated_filesystem():
        PassTheSalt().to_path('passthesalt')

        result = runner.invoke(
            cli,
            ['--path', 'passthesalt', 'rm', '--regex'],
            input='Example\n'
        )
        assert result.exit_code == 1
        assert "Error: unable to resolve pattern 'Example'" in result.output


def test_pts_rm_multiple():
    runner = CliRunner()

    with runner.isolated_filesystem():
        pts = PassTheSalt()
        pts.add('Example1', Generatable(salt='salt'))
        pts.add('Example2', Generatable(salt='salt'))
        pts.to_path('passthesalt')

        result = runner.invoke(
            cli,
            ['--path', 'passthesalt', 'rm', '--regex'],
            input='Example\ny\n'
        )
        assert result.exit_code == 0
        assert "Removed 'Example1', 'Example2'!" in result.output


def test_pts_mv():
    runner = CliRunner()

    with runner.isolated_filesystem():
        pts = PassTheSalt()
        pts.add('Example', Generatable(salt='salt'))
        pts.to_path('passthesalt')

        result = runner.invoke(
            cli,
            ['--path', 'passthesalt', 'mv', 'Example', 'Example2']
        )
        assert result.exit_code == 0
        assert "Renamed 'Example' as 'Example2'!" in result.output


def test_pts_diff():
    runner = CliRunner()

    with runner.isolated_filesystem():
        pts = PassTheSalt()
        pts.to_path('passthesalt_other')
        pts.add('Example', Generatable(salt='salt'))
        pts.to_path('passthesalt')

        result = runner.invoke(
            cli,
            ['--path', 'passthesalt', 'diff', '--path', 'passthesalt_other']
        )
        assert result.exit_code == 1
        assert 'Local store has the following extra/modified secrets:\nExample' in result.output

        result = runner.invoke(
            cli,
            ['--path', 'passthesalt_other', 'diff', '--path', 'passthesalt']
        )
        assert result.exit_code == 1
        assert 'Remote store has the following extra/modified secrets:\nExample' in result.output



"""
The core PassTheSalt module.
"""

import re
from hmac import compare_digest

import serde
from serde import Model, fields

from passthesalt.crypto import decrypt, encrypt, generate, pbkdf2_hash
from passthesalt.exceptions import ConfigurationError, ContextError, LabelError
from passthesalt.model import ModifiedModel


def version():
    """
    Return the current PassTheSalt version.

    Returns:
        str: the version.
    """
    from passthesalt import __version__
    return __version__


class Secret(ModifiedModel):
    """
    A base class for a secret.
    """

    @classmethod
    def from_dict(cls, d, strict=True):
        """
        Convert a dictionary to a Secret.

        Args:
            d (dict): a serialized version of this Model.
            strict (bool): if set to False then no exception will be raised when
                unknown dictionary keys are present.

        Returns:
            Secret: an instance of Secret.

        Raises:
            `~serde.exceptions.DeserializationError`: when the Secret kind is
                not valid.
        """
        if cls is Secret:
            kind = d.pop('kind', None)

            try:
                return SECRETS[kind].from_dict(d, strict=strict)
            except KeyError:
                raise serde.exceptions.DeserializationError(f'{kind!r} is not a valid Secret kind')

        return super().from_dict(d, strict=strict)

    def to_dict(self, dict=None):
        """
        Convert this Secret to a dictionary.

        Args:
            dict (type): the class of the deserialized dictionary. This defaults
                to an `OrderedDict` so that the fields will be returned in the
                order they were defined on the Model.

        Returns:
            dict: the Secret serialized as a dictionary.

        Raises:
            `~serde.exceptions.SerializationError`: when the Secret has an
                unknown kind.
        """
        d = super().to_dict()

        try:
            d['kind'] = KINDS[self.__class__]
        except KeyError:
            raise serde.exceptions.SerializationError(f'{self.__class__} has an unknown kind')

        return d

    def __getattr__(self, item):
        """
        Override the access of `_label` and `_pts` attributes.

        Raises:
            ContextError: if attributes `_label` or `_pts` are accessed outside
                of a PassTheSalt context.
        """
        if item in ('_label', '_pts'):
            raise ContextError(
                f'{self.__class__.__name__!r} secret is not in a PassTheSalt context'
            )

        return self.__getattribute__(item)

    def display(self):
        """
        A display tuple for tabulating this secret.

        Returns:
            (str, str): the label and the kind.
        """
        return (self._label, KINDS[self.__class__].split('.')[0], self.modified)

    def add_context(self, label, pts):
        """
        Set the context for this Secret.

        Args:
            label (str): the label for this Secret.
            pts (PassTheSalt): the PassTheSalt store for this Secret.
        """
        self._label = label
        self._pts = pts

    def check_context(self):
        """
        Check whether this Secret has a context.
        """
        self._pts
        self._label

    def remove_context(self):
        """
        Remove the context for this Secret.
        """
        del self._label
        del self._pts

    def add(self):
        """
        Add this Secret to the PassTheSalt store.
        """
        self.check_context()

    def get(self):
        """
        Retrieve the value for this Secret.

        Returns:
            str: the secret value.
        """
        raise NotImplementedError()

    def remove(self):
        """
        Remove this Secret from the PassTheSalt store.
        """
        self.check_context()


class Algorithm(Model):
    """
    A secret generation algorithm.
    """

    version = fields.Optional(fields.Int, default=1)
    length = fields.Optional(fields.Int)


class Generatable(Secret):
    """
    A generatable Secret.
    """

    salt = fields.Str()
    algorithm = fields.Optional(Algorithm, default=Algorithm)

    def display(self):
        """
        A display tuple for this tabulating this secret.

        Returns:
            (str, str, str): the label, the kind, and the salt.
        """
        return super().display() + (self.salt,)

    def get(self):
        """
        Generate the secret value for this Secret.

        Returns:
            str: the secret value.
        """
        return generate(
            self.salt,
            self._pts.master_key,
            version=self.algorithm.version,
            length=self.algorithm.length
        )


class Login(Generatable):
    """
    An account login Secret.
    """

    domain = fields.Domain()
    username = fields.Str()
    iteration = fields.Optional(fields.Int)

    @property
    def salt(self):
        """
        The salt for this Generatable secret.

        Returns:
            str: the salt.
        """
        return '|'.join((self.domain, self.username, str(self.iteration or 0)))


class Encrypted(Secret):
    """
    Represents and defines an encrypted Secret.
    """

    def __init__(self, secret, *args, **kwargs):
        """
        Create a new Encrypted.

        Args:
            secret (str): the secret to encrypt.
        """
        super().__init__(*args, **kwargs)
        self.secret = secret

    def _encrypt(self, secrets):
        """
        Encrypt a store with the master key.

        Args:
            secrets (dict): the store.
        """
        if not secrets:
            self._pts.secrets_encrypted = None
        else:
            self._pts.secrets_encrypted = encrypt(secrets, self._pts.master_key)

    def _decrypt(self):
        """
        Decrypt the encrypted store with the master key.

        Returns:
            dict: the encrypted store.
        """
        if not self._pts.secrets_encrypted:
            return {}

        return decrypt(self._pts.secrets_encrypted, self._pts.master_key)

    def add(self):
        """
        Add this Secret to the PassTheSalt store.
        """
        secrets = self._decrypt()
        secrets[self._label] = self.secret
        del self.secret
        self._encrypt(secrets)
        super().add()

    def get(self):
        """
        Decrypt secret for this Secret.

        Returns:
            str: the decrypted secret.
        """
        try:
            return self._decrypt()[self._label]
        except KeyError:
            raise LabelError(f'{self._label!r} does not exist in the encrypted store')

    def remove(self):
        """
        Remove this Secret from the PassTheSalt store.
        """
        secrets = self._decrypt()
        del secrets[self._label]
        self._encrypt(secrets)
        super().remove()


SECRETS = {
    'encrypted': Encrypted,
    'generatable': Generatable,
    'generatable.login': Login
}

KINDS = {cls: kind for kind, cls in SECRETS.items()}


class Master(Model):
    """
    Represents and defines a master password.
    """

    salt = fields.Str()
    hash = fields.Str()

    def __init__(self, master):
        """
        Configure master validation with the given master.

        Args:
            master (str): the master password.
        """
        salt, hash = pbkdf2_hash(master)
        super().__init__(salt=salt, hash=hash)

    def is_valid(self, master):
        """
        Check the given master with the stored hash.

        Args:
            master (str): the master password.

        Returns:
            bool: whether the master matches.
        """
        return compare_digest(self.hash, pbkdf2_hash(master, self.salt)[1])


class Config(Model):
    """
    Represents and defines config for PassTheSalt.
    """

    owner = fields.Optional(fields.Str)
    master = fields.Optional(Master)


class PassTheSalt(ModifiedModel):
    """
    An object to store and manage Secrets.

    A PassTheSalt represents and defines a deterministic password generation and
    password storage system.
    """

    config = fields.Optional(Config, default=Config)
    secrets = fields.Optional(fields.Dict(fields.Str, Secret), default=dict)
    secrets_encrypted = fields.Optional(fields.Str)
    version = fields.Optional(fields.Str, default=version)

    def __init__(self, *args, **kwargs):
        """
        Create a new PassTheSalt.
        """
        super().__init__(*args, **kwargs)
        self._master = None
        self.version = version()

    @classmethod
    def from_dict(cls, d, strict=True):
        """
        Create a PassTheSalt object from a dictionary.

        Args:
            d (dict): the input dictionary.
            strict (bool): if set to False then no exception will be raised when
                unknown dictionary keys are present.

        Returns:
            PassTheSalt: a new PassTheSalt object.
        """
        pts = super().from_dict(d, strict=strict)

        # Add the current context to each Secret.
        for label, secret in pts.secrets.items():
            secret.add_context(label, pts)

        return pts

    def save(self, dict=None, **kwargs):
        """
        Write this PassTheSalt store to the configured path.

        Args:
            dict (type): the class of the deserialized dictionary. This defaults
                to an `OrderedDict` so that the fields will be returned in the
                order they were defined on the Model.
            **kwargs: extra keyword arguments passed directly to `json.dumps()`.
        """
        self.to_path(self.path, dict=dict, **kwargs)

    def with_master(self, master):
        """
        Configure PassTheSalt with a master password.

        Args:
            master: the master password for generating and encrypting secrets.
                This can be a callback for getting the password (for example
                through user input), or the actual master password as a string.

        Returns:
            PassTheSalt: this object.
        """
        self._master = master
        return self

    def with_path(self, path):
        """
        Configure PassTheSalt with a default path.

        Args:
            path (str): the default path to read and write to.

        Returns:
            PassTheSalt: this object.
        """
        self._path = path
        return self

    @property
    def master_key(self):
        """
        Return the master key.

        This is  constructed from the master password and the configured owner.

        Returns:
            str: the master key.
        """
        if self._master is None:
            raise ConfigurationError('no master password is configured')

        if callable(self._master):
            self._master = self._master(self)

        key = []

        if self.config.owner:
            key.append(self.config.owner)

        key.append(self._master)

        return '|'.join(key)

    @property
    def path(self):
        """
        Return the configured path if it is set.

        Returns:
            str: the configured path.

        Raises:
            `ConfigurationError`: when there is no configured path.
        """
        try:
            return self._path
        except AttributeError:
            raise ConfigurationError('no default path is configured')

    def labels(self, pattern=None, ignorecase=True):
        """
        Return the list of labels for secrets.

        This list can be optionally filtered with a regex pattern.

        Args:
            pattern (str): filter labels with a regex pattern.

        Returns:
            list: a list of labels matching the given pattern and prefix.

        Raises:
            LabelError: when the given pattern is an invalid regex expression.
        """
        labels = self.secrets.keys()

        if pattern:
            try:
                regex = re.compile(pattern)
            except re.error:
                raise LabelError(f'{pattern!r} is an invalid regex expression')

            labels = filter(regex.match, labels)

        return list(labels)

    def resolve(self, pattern):
        """
        Resolve a pattern and prefix to a single label.

        Args:
            pattern (str): filter labels with a regex pattern.

        Returns:
            str: the actual label of the secret.

        Raises:
            LabelError: if the pattern does not match any labels or multiple
                labels are matched.
        """
        if self.contains(pattern):
            return pattern

        matches = self.labels(pattern=pattern)

        if len(matches) == 1:
            return matches[0]
        elif not matches:
            raise LabelError(f'unable to resolve pattern {pattern!r}')
        else:
            raise LabelError(f'pattern {pattern!r} matches multiple secrets')

    def contains(self, label):
        """
        Whether the label exists.

        Args:
            label (str): the label for the secret.

        Returns:
            bool: True if the label exists else False.
        """
        return label in self.secrets

    def add(self, label, secret):
        """
        Add a secret to PassTheSalt.

        Args:
            label (str): the label for the secret.
            secret (Secret): the secret to add.
        """
        if self.contains(label):
            raise LabelError(f'{label!r} already exists')

        secret.add_context(label, self)
        secret.add()
        self.secrets[label] = secret
        self.touch()

    def get(self, label):
        """
        Retrieve a secret.

        Args:
            label (str): the label for the secret.

        Returns:
            Secret: the secret corresponding to the label.
        """
        return self.secrets[label]

    def pop(self, label):
        """
        Remove a secret and return the removed secret.

        Args:
            label (str): the label for the secret.

        Returns:
            Secret: the secret corresponding to the label.
        """
        try:
            secret = self.secrets.pop(label)
        except KeyError:
            raise LabelError(f'{label!r} does not exist')

        secret.remove()
        secret.remove_context()
        self.touch()

        return secret

    def remove(self, label):
        """
        Remove a secret.

        Args:
            label (str): the label for the secret.
        """
        self.pop(label)

    def move(self, label, new_label):
        """
        Rename a secret.

        Args:
            label (str): the label for the secret.
            new_label (str): the new label for the secret.
        """
        if self.contains(new_label):
            raise LabelError(f'{new_label!r} already exists')

        self.add(new_label, self.pop(label))

    def update(self, label, secret):
        """
        Update secret.

        Args:
            label (str): the label for the secret.
            secret (Secret): the secret to update with.
        """
        if self.contains(label):
            self.remove(label)

        self.add(label, secret)

    def _diff(self, other):
        """
        Return the difference between this store and the other.

        The returned PassTheSalt store contains everything in the current store
        that is not present in the other, and anything that is not equal.

        Warning: this method is private API for a reason. The returned
        PassTheSalt store is not usable as a store.

        Args:
            other (PassTheSalt): the store to compare with.

        Returns:
            PassTheSalt: a store with all the extra / missing secrets.
        """
        diff = PassTheSalt()

        for label in self.labels():
            if label not in other.labels() or self.get(label) != other.get(label):
                diff.secrets[label] = self.get(label)

        return diff



"""
Setup file for PassTheSalt.
"""

import io
import os
import re

from setuptools import find_packages, setup


def get_metadata():
    """
    Return metadata for PassTheSalt.
    """
    here = os.path.abspath(os.path.dirname(__file__))
    init_path = os.path.join(here, 'src', 'passthesalt', '__init__.py')
    readme_path = os.path.join(here, 'README.rst')

    with io.open(init_path, encoding='utf-8') as f:
        about_text = f.read()

    metadata = {
        key: re.search(r'__' + key + r'__ = ["\'](.*?)["\']', about_text).group(1)
        for key in ('title', 'version', 'url', 'author', 'author_email', 'license', 'description')
    }
    metadata['name'] = metadata.pop('title')

    with io.open(readme_path, encoding='utf-8') as f:
        metadata['long_description'] = f.read()

    return metadata


metadata = get_metadata()

# Primary requirements
install_requires = [
    'click >=7.0, <8.0',
    'cryptography >=2.0.0, <3.0.0',
    'pyperclip >=1.0.0, <2.0.0',
    'requests >=2.0.0, <3.0.0',
    'serde[ext,toml] >=0.5.1, <0.6.0',
    'tabulate >=0.8.0, <0.9.0'
]
entry_points = {
    'console_scripts': [
        'pts=passthesalt.cli:cli',
        'passthesalt=passthesalt.cli:cli'
    ]
}

# Development requirements
lint_requires = [
    'flake8 >=3.7.0',
    'flake8-docstrings',
    'flake8-isort',
    'flake8-quotes',
    'pep8-naming'
]
test_requires = [
    'pytest >=3.6.0',
    'pytest-cov'
]

setup(
    # Options
    install_requires=install_requires,
    extras_require={
        'dev.lint': lint_requires,
        'dev.test': test_requires
    },
    python_requires='>=3.6',
    packages=find_packages('src'),
    entry_points=entry_points,
    package_dir={'': 'src'},
    py_modules=['passthesalt'],

    # Metadata
    download_url='{url}/archive/{version}.tar.gz'.format(**metadata),
    project_urls={
        'Issue Tracker': '{url}/issues'.format(**metadata)
    },
    classifiers=[
        'Environment :: Console',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython'
    ],
    keywords='password manager pbkdf2',
    **metadata
)



import datetime
import json
from unittest.mock import Mock, patch

import requests
import requests.exceptions
from pytest import raises

from passthesalt.core import PassTheSalt
from passthesalt.exceptions import (
    ConflictingTimestamps, RemoteError, UnauthorizedAccess, UnexpectedStatusCode
)
from passthesalt.remote import Remote, Stow


class TestRemote:

    def test_with_auth(self):

        def get_auth(remote):
            return ('name', 'password')

        remote = Remote(location='https://api.store.com').with_auth(get_auth)

        assert remote._auth == get_auth
        assert remote.auth == ('name', 'password')
        assert remote._auth == ('name', 'password')

    def test_validate_response(self):
        remote = Remote(location='https://api.store.com')
        response = Mock(**{
            'status_code': 200,
            'raise_for_status.return_value': None
        })
        remote.validate_response(response)

        with raises(UnexpectedStatusCode) as e:
            response = Mock(**{
                'status_code': 400,
                'raise_for_status.side_effect': requests.exceptions.HTTPError
            })
            remote.validate_response(response)
            assert e.code == 400

    def test_request(self):
        remote = Remote(location='https://api.store.com')

        with patch('passthesalt.remote.requests.request') as mock_request:
            resp = Mock(**{
                'status_code': 200,
                'json.return_value': {'message': 'result'}
            })
            mock_request.return_value = resp
            assert remote.request('GET', remote.location) == resp

            resp = Mock(**{
                'status_code': 500,
                'raise_for_status.side_effect': requests.exceptions.HTTPError,
                'json.return_value': {'message': 'result'}
            })
            mock_request.return_value = resp
            with raises(RemoteError):
                remote.request('GET', remote.location)

            resp = Mock(**{
                'status_code': 200,
                'json.side_effect': json.decoder.JSONDecodeError('msg', 'doc', 0)
            })
            mock_request.return_value = resp
            assert remote.request('GET', remote.location) == resp

            mock_request.status_code = 200
            mock_request.side_effect = requests.exceptions.RequestException
            with raises(RemoteError):
                remote.request('GET', remote.location)

    def test_get(self):
        remote = Remote(location='https://api.store.com').with_auth(('name', 'password'))

        with patch('passthesalt.remote.Remote.request') as mock_request:
            mock_request.return_value = 'eyJtb2RpZmllZCI6ICIyMDE4LTEyLTI1IDAwOjAwOjAwIn0='
            expected = PassTheSalt(
                modified=datetime.datetime(year=2018, month=12, day=25)
            )
            assert remote.get() == expected

    def test_put(self):
        remote = Remote(location='https://api.store.com').with_auth(('name', 'password'))

        with patch('passthesalt.remote.Remote.request') as mock_request:
            pts = PassTheSalt(
                modified=datetime.datetime(year=2018, month=12, day=25)
            )
            expected = (
                'eyJtb2RpZmllZCI6ICIyMDE4LTEyLTI1IiwgImNvbmZpZyI6IH'
                't9LCAic2VjcmV0cyI6IHt9LCAidmVyc2lvbiI6ICIzLjAuMCJ9'
            )
            assert remote.put(pts) == mock_request.return_value
            assert mock_request.call_args[1]['data'] == expected


class TestStow:

    def test___init__(self):
        remote = Stow(
            location='https://api.store.com',
            token_location='https://api.store.com/token'
        )
        assert remote.location == 'https://api.store.com'
        assert remote.token_location == 'https://api.store.com/token'
        assert remote.token is None

    def test_validate_response(self):
        remote = Stow(
            location='https://api.store.com',
            token_location='https://api.store.com/token'
        )
        response = Mock(**{
            'status_code': 200,
            'json.return_value': {'message': 'nuff said'},
            'raise_for_status.return_value': None
        })
        remote.validate_response(response)

        with raises(UnexpectedStatusCode) as e:
            response = Mock(**{
                'status_code': 400,
                'json.return_value': {'message': 'nuff said'},
                'raise_for_status.side_effect': requests.exceptions.HTTPError
            })
            remote.validate_response(response)
            assert e.code == 400

        with raises(UnauthorizedAccess) as e:
            response.status_code = 401
            remote.validate_response(response)
            assert e.code == 401
            assert e.message == 'nuff said'

        with raises(ConflictingTimestamps) as e:
            response.status_code = 409
            remote.validate_response(response)
            assert e.code == 409
            assert e.message == 'nuff said'

        with raises(UnexpectedStatusCode) as e:
            response = Mock(**{
                'status_code': 500,
                'json.side_effect': json.decoder.JSONDecodeError('msg', 'doc', 0),
                'raise_for_status.side_effect': requests.exceptions.HTTPError
            })
            remote.validate_response(response)
            assert e.code == 500
            assert e.message is None

    def test_renew(self):
        remote = Stow(
            location='https://api.store.com',
            token_location='https://api.store.com/token'
        ).with_auth(('name', 'password'))
        assert remote.token is None

        with patch('passthesalt.remote.requests.request') as mock_request:
            mock_request.return_value = Mock(**{
                'status_code': 200,
                'json.return_value': {'token': 'thetoken'}
            })
            assert remote.renew() is None
            assert remote.token == 'thetoken'

    def test_handle_renew(self):
        remote = Stow(
            location='https://api.store.com',
            token_location='https://api.store.com/token'
        ).with_auth(('name', 'password'))

        with patch('passthesalt.remote.requests.request') as mock_request:
            mock_request.return_value = Mock(**{
                'status_code': 200,
                'json.return_value': {'token': 'thetoken'}
            })

            decorated_function = remote.handle_renew(
                Mock(side_effect=[UnauthorizedAccess('msg'), 1])
            )
            assert decorated_function() == 1

            decorated_function = remote.handle_renew(
                Mock(side_effect=[UnexpectedStatusCode('msg', 500), 1])
            )
            with raises(UnexpectedStatusCode):
                decorated_function()

    def test_get(self):
        remote = Stow(
            location='https://api.store.com',
            token_location='https://api.store.com/token'
        ).with_auth(('name', 'password'))
        remote.token = 'thetoken'

        with patch('passthesalt.remote.Stow.request') as mock_request:
            mock_request.return_value = Mock(**{
                'status_code': 200,
                'json.return_value': {'value': 'eyJtb2RpZmllZCI6ICIyMDE4LTEyLTI1IDAwOjAwOjAwIn0='}
            })
            expected = PassTheSalt(
                modified=datetime.datetime(year=2018, month=12, day=25)
            )
            assert remote.get() == expected
            assert mock_request.call_args[1]['auth'] == ('thetoken', 'unused')

    def test_put(self):
        remote = Stow(
            location='https://api.store.com',
            token_location='https://api.store.com/token'
        )
        remote.token = 'thetoken'

        with patch('passthesalt.remote.Stow.request') as mock_request:
            mock_request.return_value = Mock(**{
                'status_code': 200,
                'json.return_value': {'message': 'success'}
            })

            pts = PassTheSalt()
            assert remote.put(pts) == 'success'
            assert mock_request.call_args[1]['auth'] == ('thetoken', 'unused')

            mock_request.return_value = Mock(**{
                'status_code': 200,
                'json.side_effect': json.decoder.JSONDecodeError('msg', 'doc', 0),
            })
            pts = PassTheSalt()
            assert remote.put(pts) is None



"""
Extensions for serde for use in PassTheSalt.
"""

import datetime
from base64 import b64decode, b64encode

from serde import Model, fields


class DateTime(fields.DateTime):
    """
    A custom DateTime Field that uses multiple valid formats.
    """

    formats = ('%Y-%m-%d %H:%M:%S.%f', '%Y-%m-%d %H:%M:%S', '%Y-%m-%d')

    def __init__(self, **kwargs):
        """
        Create a new DateTime.

        Args:
            **kwargs: keyword arguments for the `Field` constructor.
        """
        super().__init__(**kwargs)
        del self.format

    def serialize(self, value):
        """
        Serialize the given `~datetime.datetime` as a string.

        Args:
            value (~datetime.datetime): the datetime object to serialize.

        Returns:
            str: a string representation of the datetime.
        """
        if value.microsecond:
            format = self.formats[0]
        elif value.hour or value.minute or value.second:
            format = self.formats[1]
        else:
            format = self.formats[2]

        return value.strftime(format)

    def deserialize(self, value):
        """
        Deserialize the given string as a `~datetime.datetime`.

        Deserializes using one of the valid formats.

        Args:
            value (str): the string to deserialize.

        Returns:
            ~datetime.datetime: the deserialized datetime.
        """
        for datetime_format in self.formats:
            try:
                return datetime.datetime.strptime(value, datetime_format)
            except ValueError:
                pass

        raise ValueError(f'datetime {value!r} does not match any valid formats')


class ModifiedModel(Model):
    """
    A custom Model that has a modified Field.
    """

    modified = fields.Optional(DateTime, default=datetime.datetime.utcnow)

    @classmethod
    def from_base64(cls, s, strict=True, **kwargs):
        """
        Create a PassTheSalt from a base64 encoded string.

        Args:
            s (str): the base64 encoded input string.
            strict (bool): if set to False then no exception will be raised when
                unknown dictionary keys are present.
            **kwargs: extra keyword arguments passed directly to `json.loads()`.

        Returns:
            PassTheSalt: a new PassTheSalt instance.
        """
        return cls.from_json(b64decode(s.encode()).decode('utf-8'), strict=strict, **kwargs)

    @classmethod
    def from_path(cls, p, strict=True, **kwargs):
        """
        Create a PassTheSalt from the given file path.

        Args:
            p (str): the file path to read from.
            strict (bool): if set to False then no exception will be raised when
                unknown dictionary keys are present.
            **kwargs: extra keyword arguments passed directly to `json.loads()`.

        Returns:
            PassTheSalt: a new PassTheSalt instance.
        """
        with open(p) as f:
            return cls.from_json(f.read(), strict=strict, **kwargs)

    def to_base64(self, dict=None, **kwargs):
        """
        Base64 encode a JSON dumped representation of this PassTheSalt.

        Args:
            dict (type): the class of the deserialized dictionary. This defaults
                to an `OrderedDict` so that the fields will be returned in the
                order they were defined on the Model.
            **kwargs: extra keyword arguments passed directly to `json.dumps()`.

        Returns:
            str: the base64 encoded representation of the PassTheSalt store.
        """
        return b64encode(self.to_json(dict=dict, **kwargs).encode()).decode('ascii')

    def to_path(self, p, dict=None, **kwargs):
        """
        Write this PassTheSalt store to the given path.

        Args:
            p (str): the file path to write to
            dict (type): the class of the deserialized dictionary. This defaults
                to an `OrderedDict` so that the fields will be returned in the
                order they were defined on the Model.
            **kwargs: extra keyword arguments passed directly to `json.dumps()`.
        """
        with open(p, 'w') as f:
            f.write(self.to_json(dict=dict, **kwargs))

    def touch(self):
        """
        Update the modified time of this object to the current UTC time.
        """
        self.modified = datetime.datetime.utcnow()
